/*
   ******************************************************************************

   LIBGCP: color conversion

   Copyright (C) 2000 Christoph Egger   [Christoph_Egger@t-online.de]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   ******************************************************************************
 */

#include <ggi/internal/gcp.h>
#include <ggi/internal/color.h>

#include <math.h>



/*  Converts from RGBA colorspace to YUV values.
 */
int gcpRGBA2YUV(const gcp_RGBAcolor *rgba, gcp_YUVcolor *yuv)
{
	ggi_float r,g,b;


/* Formula:
 *
 * Y =  0.299 * Red + 0.587 * Green + 0.114 * Blue
 * U = -0.147 * Red - 0.289 * Green + 0.436 * Blue
 * V =  0.615 * Red - 0.515 * Green - 0.100 * Blue
 */

	r = (ggi_float)rgba->r / 65535.0f;
	g = (ggi_float)rgba->g / 65535.0f;
	b = (ggi_float)rgba->b / 65535.0f;

	yuv->y =  0.299f * r + 0.587f * g + 0.114f * b;	
	yuv->u = -0.147f * r - 0.289f * g + 0.436f * b;
	yuv->v =  0.615f * r - 0.515f * g - 0.100f * b;


	return 0;
}	/* gcpRGBA2YUV */


/*  Converts from YUV colorspace to RGBA values.
 */
int gcpYUV2RGBA(const gcp_YUVcolor *yuv, gcp_RGBAcolor *rgba)
{

	ggi_float r,g,b;

/* Formula:
 *
 * Red   = Y + 0.000f * U + 1.140f * V
 * Green = Y - 0.396f * U - 0.581f * V
 * Blue  = Y + 2.029f * U + 0.000f * V
 */

	r = yuv->y + 0.000f * yuv->u + 1.140f * yuv->v;
	g = yuv->y - 0.396f * yuv->u - 0.581f * yuv->v;
	b = yuv->y + 2.029f * yuv->u + 0.000f * yuv->v;


	r *= 65535.0f;
	g *= 65535.0f;
	b *= 65535.0f;

	rgba->r = fabs(r);
	rgba->g = fabs(g);
	rgba->b = fabs(b);
	rgba->a = 0xFFFF;


	return 0;
}	/* gcpYUV2RGBA */




/*  Converts from RGBA colorspace to CMYK values.
 */
int gcpRGBA2CMYK(const gcp_RGBAcolor *rgba, gcp_CMYKcolor *cmyk)
{


/* Formula:
 *
 * Black = MIN(1 - Red, 1 - Green, 1 - Blue);
 *
 * Cyan = (1 - Red - Black) / (1 - Black)
 * Magenta = (1 - Green - Black) / (1 - Black)
 * Yellow = (1 - Blue - Black) / (1 - Black)
 */


	cmyk->k = MIN(MIN(0xFFFF - rgba->r, 0xFFFF - rgba->g), 0xFFFF - rgba->b);

	if (cmyk->k != 0xFFFF) {
		cmyk->c = (0xFFFF - rgba->r - cmyk->k) / (0xFFFF - cmyk->k);
		cmyk->m = (0xFFFF - rgba->g - cmyk->k) / (0xFFFF - cmyk->k);
		cmyk->y = (0xFFFF - rgba->b - cmyk->k) / (0xFFFF - cmyk->k);
	} else {
		cmyk->c = 0x0000;
		cmyk->m = 0x0000;
		cmyk->y = 0x0000;
	}	/* if */

	return 0;
}	/* gcpRGBA2CMYK */


/*  Converts from CMYK colorspace to RGBA values.
 */
int gcpCMYK2RGBA(const gcp_CMYKcolor *cmyk, gcp_RGBAcolor *rgba)
{

/* Formula:
 *
 * Red   = 1 - MIN(1, Cyan * (1 - Black) + Black)
 * Green = 1 - MIN(1, Magenta * (1 - Black) + Black)
 * Blue  = 1 - MIN(1, Yellow * (1 - Black) + Black)
 */

	rgba->r = 0xFFFF - MIN(0xFFFF, cmyk->c * (0xFFFF - cmyk->k) + cmyk->k);
	rgba->g = 0xFFFF - MIN(0xFFFF, cmyk->m * (0xFFFF - cmyk->k) + cmyk->k);
	rgba->b = 0xFFFF - MIN(0xFFFF, cmyk->y * (0xFFFF - cmyk->k) + cmyk->k);
	rgba->a = 0xFFFF;


	return 0;
}	/* gcpCMYK2RGBA */



/*  Converts from HSV colorspace to RGBA values.
 */
int gcpHSV2RGBA(const gcp_HSVcolor *hsv, gcp_RGBAcolor *rgba)
{
	ggi_float f, x, y, z;
	ggi_float h, v;

	int i;


	h = (hsv->h * 180.0f / M_PI);
	v = hsv->v;

	v *= 65535.0f;

	if (hsv->s == 0.0f) {
		rgba->r = rgba->g = rgba->b = (int)v;
	} else {
		while (h < 0) {
			h += 360.0f;
		}	/* while */

		h = fmod(h, 360.0f) / 60.0f;
		i = (int)h;
		f = h - i;
		x = v * (1.0f - hsv->s);
		y = v * (1.0f - (hsv->s * f));
		z = v * (1.0f - (hsv->s * (1.0f - f)));

		switch (i) {
		case 0:
			rgba->r = v;
			rgba->g = z;
			rgba->b = x;
			break;

		case 1:
			rgba->r = y;
			rgba->g = v;
			rgba->b = x;
			break;

		case 2:
			rgba->r = x;
			rgba->g = v;
			rgba->b = z;
			break;

		case 3:
			rgba->r = x;
			rgba->g = y;
			rgba->b = v;
			break;

		case 4:
			rgba->r = z;
			rgba->g = x;
			rgba->b = v;
			break;

		case 5:
			rgba->r = v;
			rgba->g = x;
			rgba->b = y;
			break;
		}	/* switch */
	}	/* if */

	rgba->a = 0xFFFF;

	return 0;

}	/* gcpHSV2RGBA */


/*  Converts an RGBA value into the HSV colorspace.
 */
int gcpRGBA2HSV(const gcp_RGBAcolor *rgba, gcp_HSVcolor *hsv)
{
	ggi_float min, max, delta, rc, gc, bc;

	rc = (ggi_float)rgba->r / 65535.0f;
	gc = (ggi_float)rgba->g / 65535.0f;
	bc = (ggi_float)rgba->b / 65535.0f;

	max = MAX(rc, MAX(gc, bc));
	min = MIN(rc, MIN(gc, bc));
	delta = max - min;
	hsv->v = max;

	if (max != 0.0f) {
		hsv->s = delta / max;
	} else {
		hsv->s = 0.0f;
	}	/* if */

	if (hsv->s == 0.0f) {
		hsv->h = 0.0f;
	} else {
		if (rc == max) {
			hsv->h = (gc - bc) / delta;
		} else if (gc == max) {
			hsv->h = 2 + (bc - rc) / delta;
		} else if (bc == max) {
			hsv->h = 4 + (rc - gc) / delta;
		}	/* if */

		hsv->h *= 60.0f;
		if (hsv->h < 0) {
			hsv->h += 360.0f;
		}	/* if */
	}	/* if */

	hsv->h = (hsv->h / 180.0f * M_PI);

	return 0;

}	/* gcpRGBA2HSV */
